---- MODULE MemoirHLTypeSafety ----

(***************************************************************************)
(* This is a very simple proof that shows the high-level spec to be        *)
(* type-safe.                                                              *)
(***************************************************************************)

EXTENDS MemoirHLSpecification

THEOREM HLTypeSafe == HLSpec => []HLTypeInvariant \*`^\label{THEOREM-HLTypeSafe}^'
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for an Inv1-style      *)
  (* proof.  First, we prove that the initial state satisfies              *)
  (* HLTypeInvariant.  Second, we prove that the HLNext predicate          *)
  (* inductively preserves HLTypeInvariant.  Third, we use temporal        *)
  (* induction to prove that these two conditions satisfy type safety over *)
  (* all behaviors.                                                        *)
  (*************************************************************************)
  <1>1. HLInit => HLTypeInvariant
    (***********************************************************************)
    (* The base case follows trivially from the definition of HLInit and   *)
    (* the assumption that the developer-supplied constants are type-safe. *)
    (***********************************************************************)
    <2>1. HAVE HLInit
    <2>2. QED
      BY <2>1, ConstantsTypeSafe DEF HLInit, HLTypeInvariant
  <1>2. HLTypeInvariant /\ [HLNext]_HLVars => HLTypeInvariant'
    (***********************************************************************)
    (* The induction step is also fairly trivial.  We assume the           *)
    (* antecedents of the implication, then show that the consequent holds *)
    (* for both HLNext actions.                                            *)
    (***********************************************************************)
    <2>1. HAVE HLTypeInvariant /\ [HLNext]_HLVars
    <2>2. CASE UNCHANGED HLVars
      (*********************************************************************)
      (* Type safety is inductively trivial for a stuttering step.         *)
      (*********************************************************************)
      BY <2>1, <2>2 DEF HLTypeInvariant, HLVars
    <2>3. CASE HLMakeInputAvailable
      (*********************************************************************)
      (* Type safety is also trivial for a HLMakeInputAvailable action.    *)
      (*********************************************************************)
      BY <2>1, <2>3 DEF HLTypeInvariant, HLMakeInputAvailable
    <2>4. CASE HLAdvanceService
      (*********************************************************************)
      (* For a HLAdvanceService action, we just walk through the           *)
      (* definitions.  Type safety follows directly.                       *)
      (*********************************************************************)
      <3>1. PICK input \in HLAvailableInputs : HLAdvanceService!(input)!1
        BY DEF HLAdvanceService
      <3>2. /\ HLPublicState \in PublicStateType
            /\ HLPrivateState \in PrivateStateType
            /\ HLAvailableInputs \subseteq InputType
        BY <2>1 DEF HLTypeInvariant
      <3>3. HLAdvanceService!(input)!hlSResult \in ServiceResultType
        BY <3>1, <3>2, ServiceTypeSafe
      <3>4. /\ HLAdvanceService!(input)!hlSResult.newPublicState \in PublicStateType
            /\ HLAdvanceService!(input)!hlSResult.newPrivateState \in PrivateStateType
            /\ HLAdvanceService!(input)!hlSResult.output \in OutputType
        BY <3>3 DEF ServiceResultType
      <3>5. QED
        BY <2>1, <3>1, <3>4 DEF HLAdvanceService, HLTypeInvariant 
    <2>5. CASE HLDie
      (*********************************************************************)
      (* Type safety is also trivial for a HLDie action.                   *)
      (*********************************************************************)
      BY <2>1, <2>5, ConstantsTypeSafe DEF HLTypeInvariant, HLDie
    <2>6. QED
      BY <2>1, <2>2, <2>3, <2>4, <2>5 DEF HLNext
  <1>3. QED
    (***********************************************************************)
    (* Using the Inv1 proof rule, the base case and the induction step     *)
    (* together imply that the invariant always holds.                     *)
    (***********************************************************************)
    <2>1. HLTypeInvariant /\ [][HLNext]_HLVars => []HLTypeInvariant
      BY <1>2, Inv1
    <2>2. QED
      BY <2>1, <1>1 DEF HLSpec

====
