---- MODULE MemoirLL1TypeLemmas ----

(***************************************************************************)
(* This module states and proves sevaral lemmas that are useful for        *)
(* proving type safety.  Since type safety is an important part of the     *)
(* implementation proof, these lemmas also will be used in theorems other  *)
(* than the `^\basicname^' type-safety theorem.                            *)
(*                                                                         *)
(* The lemmas in this module are:                                          *)
(*     LL1SubtypeImplicationLemma                                          *)
(*     LL1InitDefsTypeSafeLemma                                            *)
(*     LL1PerformOperationDefsTypeSafeLemma                                *)
(*     LL1RepeatOperationDefsTypeSafeLemma                                 *)
(*     InclusionInvariantDefsTypeSafeLemma                                 *)
(*     CardinalityInvariantDefsTypeSafeLemma                               *)
(*     UniquenessInvariantDefsTypeSafeLemma                                *)
(*     LL1NVRAMHistorySummaryUncorruptedDefsTypeSafeLemma                  *)
(*     LL1RefinementDefsTypeSafeLemma                                      *)
(*     LL1RefinementPrimeDefsTypeSafeLemma                                 *)
(***************************************************************************)

EXTENDS MemoirLL1Refinement

(***************************************************************************)
(* LL1SubtypeImplicationLemma proves that when the LL1TypeInvariant holds, *)
(* the subtypes of LL1Disk, LL1RAM, and LL1NVRAM also hold.  This is       *)
(* asserted and proven for both the unprimed and primed states.            *)
(*                                                                         *)
(* The proof itself is completely trivial.  It follows directly from the   *)
(* type definitions LL1UntrustedStorageType and LL1TrustedStorageType.     *)
(***************************************************************************)

LL1SubtypeImplication == \*`^\label{DEF-LL1SubtypeImplication}^'
    LL1TypeInvariant =>
        /\  LL1Disk.publicState \in PublicStateType
        /\  LL1Disk.privateStateEnc \in PrivateStateEncType
        /\  LL1Disk.historySummary \in HashType
        /\  LL1Disk.authenticator \in MACType
        /\  LL1RAM.publicState \in PublicStateType
        /\  LL1RAM.privateStateEnc \in PrivateStateEncType
        /\  LL1RAM.historySummary \in HashType
        /\  LL1RAM.authenticator \in MACType
        /\  LL1NVRAM.historySummary \in HashType
        /\  LL1NVRAM.symmetricKey \in SymmetricKeyType

THEOREM LL1SubtypeImplicationLemma == \*`^\label{THEOREM-LL1SubtypeImplicationLemma}^'
    /\ LL1SubtypeImplication
    /\ LL1SubtypeImplication'
  <1>1. LL1SubtypeImplication
    <2>1. SUFFICES
          ASSUME LL1TypeInvariant
          PROVE
            /\  LL1Disk.publicState \in PublicStateType
            /\  LL1Disk.privateStateEnc \in PrivateStateEncType
            /\  LL1Disk.historySummary \in HashType
            /\  LL1Disk.authenticator \in MACType
            /\  LL1RAM.publicState \in PublicStateType
            /\  LL1RAM.privateStateEnc \in PrivateStateEncType
            /\  LL1RAM.historySummary \in HashType
            /\  LL1RAM.authenticator \in MACType
            /\  LL1NVRAM.historySummary \in HashType
            /\  LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY DEF LL1SubtypeImplication
    <2>2. LL1Disk \in LL1UntrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>3. LL1RAM \in LL1UntrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>4. LL1NVRAM \in LL1TrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>5. LL1Disk.publicState \in PublicStateType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>6. LL1Disk.privateStateEnc \in PrivateStateEncType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>7. LL1Disk.historySummary \in HashType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>8. LL1Disk.authenticator \in MACType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>9. LL1RAM.publicState \in PublicStateType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>10. LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>11. LL1RAM.historySummary \in HashType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>12. LL1RAM.authenticator \in MACType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>13. LL1NVRAM.historySummary \in HashType
      BY <2>4 DEF LL1TrustedStorageType
    <2>14. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <2>4 DEF LL1TrustedStorageType
    <2>15. QED
      BY <2>5, <2>6, <2>7, <2>8, <2>9, <2>10, <2>11, <2>12, <2>13, <2>14
  <1>2. LL1SubtypeImplication'
    <2>1. SUFFICES
          ASSUME LL1TypeInvariant'
          PROVE
            /\  LL1Disk.publicState' \in PublicStateType
            /\  LL1Disk.privateStateEnc' \in PrivateStateEncType
            /\  LL1Disk.historySummary' \in HashType
            /\  LL1Disk.authenticator' \in MACType
            /\  LL1RAM.publicState' \in PublicStateType
            /\  LL1RAM.privateStateEnc' \in PrivateStateEncType
            /\  LL1RAM.historySummary' \in HashType
            /\  LL1RAM.authenticator' \in MACType
            /\  LL1NVRAM.historySummary' \in HashType
            /\  LL1NVRAM.symmetricKey' \in SymmetricKeyType
      BY DEF LL1SubtypeImplication
    <2>2. LL1Disk' \in LL1UntrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>3. LL1RAM' \in LL1UntrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>4. LL1NVRAM' \in LL1TrustedStorageType
      BY <2>1 DEF LL1TypeInvariant
    <2>5. LL1Disk.publicState' \in PublicStateType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>6. LL1Disk.privateStateEnc' \in PrivateStateEncType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>7. LL1Disk.historySummary' \in HashType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>8. LL1Disk.authenticator' \in MACType
      BY <2>2 DEF LL1UntrustedStorageType
    <2>9. LL1RAM.publicState' \in PublicStateType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>10. LL1RAM.privateStateEnc' \in PrivateStateEncType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>11. LL1RAM.historySummary' \in HashType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>12. LL1RAM.authenticator' \in MACType
      BY <2>3 DEF LL1UntrustedStorageType
    <2>13. LL1NVRAM.historySummary' \in HashType
      BY <2>4 DEF LL1TrustedStorageType
    <2>14. LL1NVRAM.symmetricKey' \in SymmetricKeyType
      BY <2>4 DEF LL1TrustedStorageType
    <2>15. QED
      BY <2>5, <2>6, <2>7, <2>8, <2>9, <2>10, <2>11, <2>12, <2>13, <2>14
  <1>3. QED
    BY <1>1, <1>2

(***************************************************************************)
(* LL1InitDefsTypeSafeLemma proves that the definitions within the LET of  *)
(* the LL1Init action all have the appropriate type.  This is a trivial    *)
(* proof that merely walks through the definitions.                        *)
(***************************************************************************)

THEOREM LL1InitDefsTypeSafeLemma == \*`^\label{THEOREM-LL1InitDefsTypeSafeLemma}^'
    \A symmetricKey \in SymmetricKeyType :
        LET
            initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
            initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
            initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
            initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
            initialUntrustedStorage == [
                publicState |-> InitialPublicState,
                privateStateEnc |-> initialPrivateStateEnc,
                historySummary |-> BaseHashValue,
                authenticator |-> initialAuthenticator ]
            initialTrustedStorage == [
                historySummary |-> BaseHashValue,
                symmetricKey |-> symmetricKey ]
        IN
            /\  initialPrivateStateEnc \in PrivateStateEncType
            /\  initialStateHash \in HashType
            /\  initialHistoryStateBinding \in HashType
            /\  initialAuthenticator \in MACType
            /\  initialUntrustedStorage \in LL1UntrustedStorageType
            /\  initialTrustedStorage \in LL1TrustedStorageType
  <1>1. TAKE symmetricKey \in SymmetricKeyType
  <1> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
  <1> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
  <1> initialHistoryStateBinding == Hash(BaseHashValue, initialStateHash)
  <1> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
  <1> initialUntrustedStorage == [
          publicState |-> InitialPublicState,
          privateStateEnc |-> initialPrivateStateEnc,
          historySummary |-> BaseHashValue,
          authenticator |-> initialAuthenticator ]
  <1> initialTrustedStorage == [
          historySummary |-> BaseHashValue,
          symmetricKey |-> symmetricKey ]
  <1> HIDE DEF initialPrivateStateEnc, initialStateHash, initialAuthenticator,
          initialUntrustedStorage, initialTrustedStorage
  <1>2. initialPrivateStateEnc \in PrivateStateEncType
    <2>1. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>2. InitialPrivateState \in PrivateStateType
      BY ConstantsTypeSafe
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF initialPrivateStateEnc
  <1>3. initialStateHash \in HashType
    <2>1. InitialPublicState \in HashDomain
      <3>1. InitialPublicState \in PublicStateType
        BY ConstantsTypeSafe
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. initialPrivateStateEnc \in HashDomain
      BY <1>2 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF initialStateHash
  <1>4. initialHistoryStateBinding \in HashType
    <2>1. BaseHashValue \in HashDomain
      <3>1. BaseHashValue \in HashType
        BY BaseHashValueTypeSafe
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. initialStateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF initialHistoryStateBinding
  <1>5. initialAuthenticator \in MACType
    <2>1. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>2. initialHistoryStateBinding \in HashType
      BY <1>4
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF initialAuthenticator
  <1>6. initialUntrustedStorage \in LL1UntrustedStorageType
    <2>1. InitialPublicState \in PublicStateType
      BY ConstantsTypeSafe
    <2>2. initialPrivateStateEnc \in PrivateStateEncType
      BY <1>2
    <2>3. BaseHashValue \in HashType
      BY BaseHashValueTypeSafe
    <2>4. initialAuthenticator \in MACType
      BY <1>5
    <2>5. QED
      BY <2>1, <2>2, <2>3, <2>4 DEF initialUntrustedStorage, LL1UntrustedStorageType
  <1>7. initialTrustedStorage \in LL1TrustedStorageType
    <2>1. BaseHashValue \in HashType
      BY BaseHashValueTypeSafe
    <2>2. symmetricKey \in SymmetricKeyType
      BY <1>1
    <2>3. QED
      BY <2>1, <2>2 DEF initialTrustedStorage, LL1TrustedStorageType
  <1>8. QED
    BY <1>2, <1>3, <1>4, <1>5, <1>6, <1>7
    DEF initialPrivateStateEnc, initialStateHash, initialAuthenticator,
        initialUntrustedStorage, initialTrustedStorage

(***************************************************************************)
(* LL1PerformOperationDefsTypeSafeLemma proves that the definitions within *)
(* the LET of the LL1PerformOperation action all have the appropriate      *)
(* type.  This is a trivial proof that merely walks through the            *)
(* definitions.                                                            *)
(***************************************************************************)

THEOREM LL1PerformOperationDefsTypeSafeLemma == \*`^\label{THEOREM-LL1PerformOperationDefsTypeSafeLemma}^'
    \A input \in LL1AvailableInputs :
        LL1TypeInvariant =>
            LET
                stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
                historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
                privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
                sResult == Service(LL1RAM.publicState, privateState, input)
                newPrivateStateEnc ==
                    SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
                newHistorySummary == Hash(LL1NVRAM.historySummary, input)
                newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
                newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
                newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
            IN
                /\  stateHash \in HashType
                /\  historyStateBinding \in HashType
                /\  privateState \in PrivateStateType
                /\  sResult \in ServiceResultType
                /\  sResult.newPublicState \in PublicStateType
                /\  sResult.newPrivateState \in PrivateStateType
                /\  sResult.output \in OutputType
                /\  newPrivateStateEnc \in PrivateStateEncType
                /\  newHistorySummary \in HashType
                /\  newStateHash \in HashType
                /\  newHistoryStateBinding \in HashType
                /\  newAuthenticator \in MACType
  <1>1. TAKE input \in LL1AvailableInputs
  <1> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
  <1> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
  <1> sResult == Service(LL1RAM.publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
  <1> newHistorySummary == Hash(LL1NVRAM.historySummary, input)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(newHistorySummary, newStateHash)
  <1> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
          newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator
  <1>2. HAVE LL1TypeInvariant
  <1>3. stateHash \in HashType
    <2>1. /\  LL1RAM.publicState \in PublicStateType
          /\  LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. /\  LL1RAM.publicState \in HashDomain
          /\  LL1RAM.privateStateEnc \in HashDomain
      BY <2>1 DEF HashDomain
    <2>3. QED
      BY <2>2, HashTypeSafe DEF stateHash
  <1>4. historyStateBinding \in HashType
    <2>1. LL1RAM.historySummary \in HashDomain
      <3>1. LL1RAM.historySummary \in HashType
        BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding
  <1>5. privateState \in PrivateStateType
    <2>1. /\  LL1NVRAM.symmetricKey \in SymmetricKeyType
          /\  LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. QED
      BY <2>1, SymmetricDecryptionTypeSafe DEF privateState
  <1>6. sResult \in ServiceResultType
    <2>1. LL1RAM.publicState \in PublicStateType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. privateState \in PrivateStateType
      BY <1>5
    <2>3. input \in InputType
      <3>1. LL1AvailableInputs \subseteq InputType
        BY <1>2 DEF LL1TypeInvariant
      <3>2. QED
        BY <1>1, <3>1
    <2>4. QED
      BY <2>1, <2>2, <2>3, ServiceTypeSafe DEF sResult
  <1>7. /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
    BY <1>6 DEF ServiceResultType
  <1>8. newPrivateStateEnc \in PrivateStateEncType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. sResult.newPrivateState \in PrivateStateType
      BY <1>7
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF newPrivateStateEnc
  <1>9. newHistorySummary \in HashType
    <2>1. LL1NVRAM.historySummary \in HashDomain
      <3>1. LL1NVRAM.historySummary \in HashType
        BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. input \in HashDomain
      <3>1. input \in InputType
        <4>1. LL1AvailableInputs \subseteq InputType
          BY <1>2 DEF LL1TypeInvariant
        <4>2. QED
          BY <1>1, <4>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistorySummary
  <1>10. newStateHash \in HashType
    <2>1. sResult.newPublicState \in HashDomain
      <3>1. sResult.newPublicState \in PublicStateType
        BY <1>7
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newPrivateStateEnc \in HashDomain
      BY <1>8 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newStateHash
  <1>11. newHistoryStateBinding \in HashType
    <2>1. newHistorySummary \in HashDomain
      BY <1>9 DEF HashDomain
    <2>2. newStateHash \in HashDomain
      BY <1>10 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistoryStateBinding
  <1>12. newAuthenticator \in MACType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. newHistoryStateBinding \in HashType
      BY <1>11
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF newAuthenticator
  <1>13. QED
    BY <1>3, <1>4, <1>5, <1>6, <1>7, <1>8, <1>9, <1>10, <1>11, <1>12
    DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
        newHistorySummary, newStateHash, newHistoryStateBinding, newAuthenticator

(***************************************************************************)
(* LL1RepeatOperationDefsTypeSafeLemma proves that the definitions within  *)
(* the LET of the LL1RepeatOperation action all have the appropriate type. *)
(* This is a trivial proof that merely walks through the definitions.      *)
(***************************************************************************)

THEOREM LL1RepeatOperationDefsTypeSafeLemma == \*`^\label{THEOREM-LL1RepeatOperationDefsTypeSafeLemma}^'
    \A input \in LL1AvailableInputs :
        LL1TypeInvariant =>
            LET
                stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
                historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
                privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
                sResult == Service(LL1RAM.publicState, privateState, input)
                newPrivateStateEnc ==
                    SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
                newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
                newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
                newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
            IN
                /\  stateHash \in HashType
                /\  historyStateBinding \in HashType
                /\  privateState \in PrivateStateType
                /\  sResult \in ServiceResultType
                /\  sResult.newPublicState \in PublicStateType
                /\  sResult.newPrivateState \in PrivateStateType
                /\  sResult.output \in OutputType
                /\  newPrivateStateEnc \in PrivateStateEncType
                /\  newStateHash \in HashType
                /\  newHistoryStateBinding \in HashType
                /\  newAuthenticator \in MACType
  <1>1. TAKE input \in LL1AvailableInputs
  <1> stateHash == Hash(LL1RAM.publicState, LL1RAM.privateStateEnc)
  <1> historyStateBinding == Hash(LL1RAM.historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, LL1RAM.privateStateEnc)
  <1> sResult == Service(LL1RAM.publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
  <1> newAuthenticator == GenerateMAC(LL1NVRAM.symmetricKey, newHistoryStateBinding)
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
          newStateHash, newHistoryStateBinding, newAuthenticator
  <1>2. HAVE LL1TypeInvariant
  <1>3. stateHash \in HashType
    <2>1. /\  LL1RAM.publicState \in PublicStateType
          /\  LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. /\  LL1RAM.publicState \in HashDomain
          /\  LL1RAM.privateStateEnc \in HashDomain
      BY <2>1 DEF HashDomain
    <2>3. QED
      BY <2>2, HashTypeSafe DEF stateHash
  <1>4. historyStateBinding \in HashType
    <2>1. LL1RAM.historySummary \in HashDomain
      <3>1. LL1RAM.historySummary \in HashType
        BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding
  <1>5. privateState \in PrivateStateType
    <2>1. /\  LL1NVRAM.symmetricKey \in SymmetricKeyType
          /\  LL1RAM.privateStateEnc \in PrivateStateEncType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. QED
      BY <2>1, SymmetricDecryptionTypeSafe DEF privateState
  <1>6. sResult \in ServiceResultType
    <2>1. LL1RAM.publicState \in PublicStateType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. privateState \in PrivateStateType
      BY <1>5
    <2>3. input \in InputType
      <3>1. LL1AvailableInputs \subseteq InputType
        BY <1>2 DEF LL1TypeInvariant
      <3>2. QED
        BY <1>1, <3>1
    <2>4. QED
      BY <2>1, <2>2, <2>3, ServiceTypeSafe DEF sResult
  <1>7. /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
    BY <1>6 DEF ServiceResultType
  <1>8. newPrivateStateEnc \in PrivateStateEncType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. sResult.newPrivateState \in PrivateStateType
      BY <1>7
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF newPrivateStateEnc
  <1>9. newStateHash \in HashType
    <2>1. sResult.newPublicState \in HashDomain
      <3>1. sResult.newPublicState \in PublicStateType
        BY <1>7
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newPrivateStateEnc \in HashDomain
      BY <1>8 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newStateHash
  <1>10. newHistoryStateBinding \in HashType
    <2>1. LL1NVRAM.historySummary \in HashDomain
      <3>1. LL1NVRAM.historySummary \in HashType
        BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newStateHash \in HashDomain
      BY <1>9 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistoryStateBinding
  <1>11. newAuthenticator \in MACType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. newHistoryStateBinding \in HashType
      BY <1>10
    <2>3. QED
      BY <2>1, <2>2, GenerateMACTypeSafe DEF newAuthenticator
  <1>12. QED
    BY <1>3, <1>4, <1>5, <1>6, <1>7, <1>8, <1>9, <1>10, <1>11
    DEF stateHash, historyStateBinding, privateState, sResult, newPrivateStateEnc,
        newStateHash, newHistoryStateBinding, newAuthenticator

----

(***************************************************************************)
(* The InclusionInvariantDefsTypeSafeLemma proves that the definitions     *)
(* within the LET of the InclusionInvariant all have the appropriate type. *)
(* This is a trivial proof that merely walks through the definitions.      *)
(***************************************************************************)

THEOREM InclusionInvariantDefsTypeSafeLemma == \*`^\label{THEOREM-InclusionInvariantDefsTypeSafeLemma}^'
    \A  input \in InputType,
        historySummary \in HashType,
        publicState \in PublicStateType,
        privateStateEnc \in PrivateStateEncType :
        LL1TypeInvariant =>
            LET
                stateHash == Hash(publicState, privateStateEnc)
                historyStateBinding == Hash(historySummary, stateHash)
                privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
                sResult == Service(publicState, privateState, input)
                newPrivateStateEnc ==
                    SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
                newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
                newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
            IN
                /\  stateHash \in HashType
                /\  historyStateBinding \in HashType
                /\  privateState \in PrivateStateType
                /\  sResult \in ServiceResultType
                /\  sResult.newPublicState \in PublicStateType
                /\  sResult.newPrivateState \in PrivateStateType
                /\  sResult.output \in OutputType
                /\  newPrivateStateEnc \in PrivateStateEncType
                /\  newStateHash \in HashType
                /\  newHistoryStateBinding \in HashType
  <1>1. TAKE input \in InputType,
             historySummary \in HashType,
             publicState \in PublicStateType,
             privateStateEnc \in PrivateStateEncType
  <1> stateHash == Hash(publicState, privateStateEnc)
  <1> historyStateBinding == Hash(historySummary, stateHash)
  <1> privateState == SymmetricDecrypt(LL1NVRAM.symmetricKey, privateStateEnc)
  <1> sResult == Service(publicState, privateState, input)
  <1> newPrivateStateEnc ==
          SymmetricEncrypt(LL1NVRAM.symmetricKey, sResult.newPrivateState)
  <1> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
  <1> newHistoryStateBinding == Hash(LL1NVRAM.historySummary, newStateHash)
  <1> HIDE DEF stateHash, historyStateBinding, privateState, sResult,
          newPrivateStateEnc, newStateHash, newHistoryStateBinding
  <1>2. HAVE LL1TypeInvariant
  <1>3. stateHash \in HashType
    <2>1. /\  publicState \in PublicStateType
          /\  privateStateEnc \in PrivateStateEncType
      BY <1>1
    <2>2. /\  publicState \in HashDomain
          /\  privateStateEnc \in HashDomain
      BY <2>1 DEF HashDomain
    <2>3. QED
      BY <2>2, HashTypeSafe DEF stateHash
  <1>4. historyStateBinding \in HashType
    <2>1. historySummary \in HashDomain
      <3>1. historySummary \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding
  <1>5. privateState \in PrivateStateType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. privateStateEnc \in PrivateStateEncType
      BY <1>1
    <2>3. QED
      BY <2>1, <2>2, SymmetricDecryptionTypeSafe DEF privateState
  <1>6. sResult \in ServiceResultType
    <2>1. publicState \in PublicStateType
      BY <1>1
    <2>2. privateState \in PrivateStateType
      BY <1>5
    <2>3. input \in InputType
      <3>1. LL1AvailableInputs \subseteq InputType
        BY <1>2 DEF LL1TypeInvariant
      <3>2. QED
        BY <1>1, <3>1
    <2>4. QED
      BY <2>1, <2>2, <2>3, ServiceTypeSafe DEF sResult
  <1>7. /\  sResult.newPublicState \in PublicStateType
        /\  sResult.newPrivateState \in PrivateStateType
        /\  sResult.output \in OutputType
    BY <1>6 DEF ServiceResultType
  <1>8. newPrivateStateEnc \in PrivateStateEncType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. sResult.newPrivateState \in PrivateStateType
      BY <1>7
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF newPrivateStateEnc
  <1>9. newStateHash \in HashType
    <2>1. sResult.newPublicState \in HashDomain
      <3>1. sResult.newPublicState \in PublicStateType
        BY <1>7
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newPrivateStateEnc \in HashDomain
      BY <1>8 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newStateHash
  <1>10. newHistoryStateBinding \in HashType
    <2>1. LL1NVRAM.historySummary \in HashDomain
      <3>1. LL1NVRAM.historySummary \in HashType
        BY <1>2, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. newStateHash \in HashDomain
      BY <1>9 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF newHistoryStateBinding
  <1>11. QED
    BY <1>3, <1>4, <1>5, <1>6, <1>7, <1>8, <1>9, <1>10
    DEF stateHash, historyStateBinding, privateState, sResult,
        newPrivateStateEnc, newStateHash, newHistoryStateBinding

----

(***************************************************************************)
(* The CardinalityInvariantDefsTypeSafeLemma proves that the definition    *)
(* within the LET of the CardinalityInvariant has the appropriate type.    *)
(* This is a trivial proof that merely walks through the definitions.      *)
(***************************************************************************)

THEOREM CardinalityInvariantDefsTypeSafeLemma == \*`^\label{THEOREM-CardinalityInvariantDefsTypeSafeLemma}^'
    \A historySummary \in HashType, stateHash \in HashType :
        LET
            historyStateBinding == Hash(historySummary, stateHash)
        IN
            historyStateBinding \in HashType
  <1>1. TAKE historySummary \in HashType, stateHash \in HashType
  <1> historyStateBinding == Hash(historySummary, stateHash)
  <1> HIDE DEF historyStateBinding
  <1>2. QED
    <2>1. historySummary \in HashDomain
      <3>1. historySummary \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. stateHash \in HashDomain
      <3>1. stateHash \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF historyStateBinding

----

(***************************************************************************)
(* The UniquenessInvariantDefsTypeSafeLemma proves that the definitions    *)
(* within the LET of the UniquenessInvariant all have the appropriate      *)
(* type.  This is a trivial proof that merely walks through the            *)
(* definitions.                                                            *)
(***************************************************************************)

THEOREM UniquenessInvariantDefsTypeSafeLemma == \*`^\label{THEOREM-UniquenessInvariantDefsTypeSafeLemma}^'
    \A  stateHash1, stateHash2 \in HashType :
        LL1TypeInvariant =>
            LET
                historyStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
                historyStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
            IN
                /\  historyStateBinding1 \in HashType
                /\  historyStateBinding2 \in HashType
  <1>1. TAKE stateHash1, stateHash2 \in HashType
  <1> historyStateBinding1 == Hash(LL1NVRAM.historySummary, stateHash1)
  <1> historyStateBinding2 == Hash(LL1NVRAM.historySummary, stateHash2)
  <1> HIDE DEF historyStateBinding1, historyStateBinding2
  <1>2. HAVE LL1TypeInvariant
  <1>3. LL1NVRAM.historySummary \in HashDomain
    <2>1. LL1NVRAM.historySummary \in HashType
      <3>1. LL1TypeInvariant
        BY <1>2
      <3>2. QED
        BY <3>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. QED
      BY <2>1 DEF HashDomain
  <1>4. historyStateBinding1 \in HashType
    <2>1. stateHash1 \in HashDomain
      <3>1. stateHash1 \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. QED
      BY <1>3, <2>1, HashTypeSafe DEF historyStateBinding1
  <1>5. historyStateBinding2 \in HashType
    <2>1. stateHash2 \in HashDomain
      <3>1. stateHash2 \in HashType
        BY <1>1
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. QED
      BY <1>3, <2>1, HashTypeSafe DEF historyStateBinding2

  <1>6. QED
    BY <1>4, <1>5
    DEF historyStateBinding1, historyStateBinding2

----

(***************************************************************************)
(* The LL1RefinementDefsTypeSafeLemma proves that the definitions within   *)
(* the LET of the LL1Refinement definition all have the appropriate type   *)
(* in the unprimed state.  This is a trivial proof that merely walks       *)
(* through the definitions.                                                *)
(***************************************************************************)

THEOREM LL1RefinementDefsTypeSafeLemma == \*`^\label{THEOREM-LL1RefinementDefsTypeSafeLemma}^'
    LL1Refinement /\ LL1TypeInvariant =>
        LET
            refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
            refStateHash == Hash(HLPublicState, refPrivateStateEnc)
            refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
        IN
            /\  refPrivateStateEnc \in PrivateStateEncType
            /\  refStateHash \in HashType
            /\  refHistoryStateBinding \in HashType
  <1> refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
  <1> refStateHash == Hash(HLPublicState, refPrivateStateEnc)
  <1> refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
  <1> HIDE DEF refPrivateStateEnc, refStateHash
  <1>1. HAVE LL1Refinement /\ LL1TypeInvariant
  <1>2. refPrivateStateEnc \in PrivateStateEncType
    <2>1. LL1NVRAM.symmetricKey \in SymmetricKeyType
      BY <1>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. HLPrivateState \in PrivateStateType
      BY <1>1, ConstantsTypeSafe DEF LL1Refinement
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF refPrivateStateEnc
  <1>3. refStateHash \in HashType
    <2>1. HLPublicState \in HashDomain
      <3>1. HLPublicState \in PublicStateType
        BY <1>1, ConstantsTypeSafe DEF LL1Refinement
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. refPrivateStateEnc \in HashDomain
      BY <1>2 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF refStateHash
  <1>4. refHistoryStateBinding \in HashType
    <2>1. LL1NVRAM.historySummary \in HashDomain
      <3>1. LL1NVRAM.historySummary \in HashType
        BY <1>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. refStateHash \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe
  <1>5. QED
    BY <1>2, <1>3, <1>4
    DEF refPrivateStateEnc, refStateHash

----

(***************************************************************************)
(* The LL1RefinementPrimeDefsTypeSafeLemma proves that the definitions     *)
(* within the LET of the LL1Refinement definition all have the appropriate *)
(* type in the primed state.  This is a trivial proof that merely walks    *)
(* through the definitions.                                                *)
(***************************************************************************)

THEOREM LL1RefinementPrimeDefsTypeSafeLemma == \*`^\label{THEOREM-LL1RefinementPrimeDefsTypeSafeLemma}^'
    LL1Refinement' /\ LL1TypeInvariant' =>
        LET
            refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
            refStateHash == Hash(HLPublicState, refPrivateStateEnc)
            refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
        IN
            /\  refPrivateStateEnc' \in PrivateStateEncType
            /\  refStateHash' \in HashType
            /\  refHistoryStateBinding' \in HashType
  <1> refPrivateStateEnc == SymmetricEncrypt(LL1NVRAM.symmetricKey, HLPrivateState)
  <1> refStateHash == Hash(HLPublicState, refPrivateStateEnc)
  <1> refHistoryStateBinding == Hash(LL1NVRAM.historySummary, refStateHash)
  <1> HIDE DEF refPrivateStateEnc, refStateHash
  <1>1. HAVE LL1Refinement' /\ LL1TypeInvariant'
  <1>2. refPrivateStateEnc' \in PrivateStateEncType
    <2>1. LL1NVRAM.symmetricKey' \in SymmetricKeyType
      BY <1>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
    <2>2. HLPrivateState' \in PrivateStateType
      BY <1>1, ConstantsTypeSafe DEF LL1Refinement
    <2>3. QED
      BY <2>1, <2>2, SymmetricEncryptionTypeSafe DEF refPrivateStateEnc
  <1>3. refStateHash' \in HashType
    <2>1. HLPublicState' \in HashDomain
      <3>1. HLPublicState' \in PublicStateType
        BY <1>1, ConstantsTypeSafe DEF LL1Refinement
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. refPrivateStateEnc' \in HashDomain
      BY <1>2 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe DEF refStateHash
  <1>4. refHistoryStateBinding' \in HashType
    <2>1. LL1NVRAM.historySummary' \in HashDomain
      <3>1. LL1NVRAM.historySummary' \in HashType
        BY <1>1, LL1SubtypeImplicationLemma DEF LL1SubtypeImplication
      <3>2. QED
        BY <3>1 DEF HashDomain
    <2>2. refStateHash' \in HashDomain
      BY <1>3 DEF HashDomain
    <2>3. QED
      BY <2>1, <2>2, HashTypeSafe
  <1>5. QED
    BY <1>2, <1>3, <1>4
    DEF refPrivateStateEnc, refStateHash

====
