---- MODULE MemoirLL2TypeSafety ----

(***************************************************************************)
(* This module proves the type safety of the `^\fastname^' spec.           *)
(***************************************************************************)

EXTENDS MemoirLL2TypeLemmas

THEOREM LL2TypeSafe == LL2Spec => []LL2TypeInvariant \*`^\label{THEOREM-LL2TypeSafe}^'
  (*************************************************************************)
  (* The top level of the proof is boilerplate TLA+ for an Inv1-style      *)
  (* proof.  First, we prove that the initial state satisfies              *)
  (* LL2TypeInvariant.  Second, we prove that the LL2Next predicate        *)
  (* inductively preserves LL2TypeInvariant.  Third, we use temporal       *)
  (* induction to prove that these two conditions satisfy type safety over *)
  (* all behaviors.                                                        *)
  (*************************************************************************)
  <1>1. LL2Init => LL2TypeInvariant
    (***********************************************************************)
    (* The base case follows directly from the definition of LL2Init.      *)
    (* There are a bunch of steps, but they are simple expansions of       *)
    (* definitions and appeals to the type safety of the initial           *)
    (* definitions.                                                        *)
    (***********************************************************************)
    <2>1. HAVE LL2Init
    <2>2. PICK symmetricKey \in SymmetricKeyType, hashBarrier \in HashType :
              LL2Init!(symmetricKey, hashBarrier)!1
      BY <2>1 DEF LL2Init
    <2> initialPrivateStateEnc == SymmetricEncrypt(symmetricKey, InitialPrivateState)
    <2> initialStateHash == Hash(InitialPublicState, initialPrivateStateEnc)
    <2> initialHistorySummary == [
            anchor |-> BaseHashValue,
            extension |-> BaseHashValue ]
    <2> initialHistorySummaryHash == Hash(BaseHashValue, BaseHashValue)
    <2> initialHistoryStateBinding == Hash(initialHistorySummaryHash, initialStateHash)
    <2> initialAuthenticator == GenerateMAC(symmetricKey, initialHistoryStateBinding)
    <2> initialUntrustedStorage == [
            publicState |-> InitialPublicState,
            privateStateEnc |-> initialPrivateStateEnc,
            historySummary |-> initialHistorySummary,
            authenticator |-> initialAuthenticator ]
    <2> initialTrustedStorage == [
            historySummaryAnchor |-> BaseHashValue,
            symmetricKey |-> symmetricKey,
            hashBarrier |-> hashBarrier,
            extensionInProgress |-> FALSE ]
    <2>3.   /\  initialPrivateStateEnc \in PrivateStateEncType
            /\  initialStateHash \in HashType
            /\  initialHistorySummary \in HistorySummaryType
            /\  initialHistorySummaryHash \in HashType
            /\  initialHistoryStateBinding \in HashType
            /\  initialAuthenticator \in MACType
            /\  initialUntrustedStorage \in LL2UntrustedStorageType
            /\  initialTrustedStorage \in LL2TrustedStorageType
      <3>1. symmetricKey \in SymmetricKeyType
        BY <2>2
      <3>2. QED
        BY <3>1, LL2InitDefsTypeSafeLemma
    <2> HIDE DEF initialPrivateStateEnc, initialStateHash, initialHistorySummary,
            initialHistorySummaryHash, initialHistoryStateBinding, initialAuthenticator,
            initialUntrustedStorage, initialTrustedStorage
    <2>4. LL2AvailableInputs \subseteq InputType
      <3>1. LL2AvailableInputs = InitialAvailableInputs
        BY <2>2
      <3>2. InitialAvailableInputs \subseteq InputType
        BY ConstantsTypeSafe DEF ConstantsTypeSafe
      <3>3. QED
        BY <3>1, <3>2
    <2>5. LL2ObservedOutputs \subseteq OutputType
      <3>1. LL2ObservedOutputs = {}
        BY <2>2
      <3>2. QED
        BY <3>1
    <2>6. LL2ObservedAuthenticators \subseteq MACType
      <3>1. LL2ObservedAuthenticators = {initialAuthenticator}
        BY <2>2
        DEF initialAuthenticator, initialHistoryStateBinding, initialHistorySummaryHash,
            initialStateHash, initialPrivateStateEnc
      <3>2. initialAuthenticator \in MACType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>7. LL2Disk \in LL2UntrustedStorageType
      <3>1. LL2Disk = initialUntrustedStorage
        BY <2>2
        DEF initialUntrustedStorage, initialHistorySummary, initialAuthenticator,
            initialHistorySummaryHash, initialHistoryStateBinding, initialStateHash, initialPrivateStateEnc
      <3>2. initialUntrustedStorage \in LL2UntrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>8. LL2RAM \in LL2UntrustedStorageType
      <3>1. LL2RAM = initialUntrustedStorage
        BY <2>2
        DEF initialUntrustedStorage, initialHistorySummary, initialAuthenticator,
            initialHistorySummaryHash, initialHistoryStateBinding, initialStateHash, initialPrivateStateEnc
      <3>2. initialUntrustedStorage \in LL2UntrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>9. LL2NVRAM \in LL2TrustedStorageType
      <3>1. LL2NVRAM = initialTrustedStorage
        BY <2>2 DEF initialTrustedStorage
      <3>2. initialTrustedStorage \in LL2TrustedStorageType
        BY <2>3
      <3>3. QED
        BY <3>1, <3>2
    <2>10. LL2SPCR \in HashType
      <3>1. LL2SPCR = BaseHashValue
        BY <2>2
      <3>2. BaseHashValue \in HashType
        BY BaseHashValueTypeSafe
      <3>3. QED
        BY <3>1, <3>2
    <2>11. QED
      BY <2>4, <2>5, <2>6, <2>7, <2>8, <2>9, <2>10 DEF LL2TypeInvariant
  <1>2. LL2TypeInvariant /\ [LL2Next]_LL2Vars => LL2TypeInvariant'
    (***********************************************************************)
    (* The induction step is also straightforward.  We assume the          *)
    (* antecedents of the implication, then show that the consequent holds *)
    (* for all nine LL2Next actions plus stuttering.                       *)
    (***********************************************************************)
    <2>1. HAVE LL2TypeInvariant /\ [LL2Next]_LL2Vars
    <2>2. CASE UNCHANGED LL2Vars
      (*********************************************************************)
      (* Type safety is inductively trivial for a stuttering step.         *)
      (*********************************************************************)
      <3>1. LL2AvailableInputs' \subseteq InputType
        <4>1. LL2AvailableInputs \subseteq InputType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2AvailableInputs
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>2. LL2ObservedOutputs' \subseteq OutputType
        <4>1. LL2ObservedOutputs \subseteq OutputType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2ObservedOutputs
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>3. LL2ObservedAuthenticators' \subseteq MACType
        <4>1. LL2ObservedAuthenticators \subseteq MACType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2ObservedAuthenticators
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>4. LL2Disk' \in LL2UntrustedStorageType
        <4>1. LL2Disk \in LL2UntrustedStorageType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2Disk
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>5. LL2RAM' \in LL2UntrustedStorageType
        <4>1. LL2RAM \in LL2UntrustedStorageType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2RAM
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>6. LL2NVRAM' \in LL2TrustedStorageType
        <4>1. LL2NVRAM \in LL2TrustedStorageType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2NVRAM
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>7. LL2SPCR' \in HashType
        <4>1. LL2SPCR \in HashType
          BY <2>1 DEF LL2TypeInvariant
        <4>2. UNCHANGED LL2SPCR
          BY <2>2 DEF LL2Vars
        <4>3. QED
          BY <4>1, <4>2
      <3>8. QED
        BY <3>1, <3>2, <3>3, <3>4, <3>5, <3>6, <3>7 DEF LL2TypeInvariant
    <2>3. CASE LL2Next
      <3>1. CASE LL2MakeInputAvailable
        (*******************************************************************)
        (* Type safety is straightforward for a LL2MakeInputAvailable      *)
        (* action.                                                         *)
        (*******************************************************************)
        <4>1. PICK input \in InputType : LL2MakeInputAvailable!(input)
          BY <3>1 DEF LL2MakeInputAvailable
        <4>2. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2AvailableInputs' = LL2AvailableInputs \union {input}
            BY <4>1
          <5>3. input \in InputType
            BY <4>1
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>3. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2RAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>9. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL2TypeInvariant
      <3>2. CASE LL2PerformOperation
        (*******************************************************************)
        (* For a LL2PerformOperation action, we just walk through the      *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4>1. PICK input \in LL2AvailableInputs : LL2PerformOperation!(input)!1
          BY <3>2 DEF LL2PerformOperation
        <4> historySummaryHash ==
                Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
        <4> stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
        <4> historyStateBinding == Hash(historySummaryHash, stateHash)
        <4> privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
        <4> sResult == Service(LL2RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
        <4> currentHistorySummary == [
                anchor |-> LL2NVRAM.historySummaryAnchor,
                extension |-> LL2SPCR ]
        <4> newHistorySummary == Successor(currentHistorySummary, input, LL2NVRAM.hashBarrier)
        <4> newHistorySummaryHash == Hash(newHistorySummary.anchor, newHistorySummary.extension)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(newHistorySummaryHash, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
        <4>2. /\  historySummaryHash \in HashType
              /\  stateHash \in HashType
              /\  historyStateBinding \in HashType
              /\  privateState \in PrivateStateType
              /\  sResult \in ServiceResultType
              /\  sResult.newPublicState \in PublicStateType
              /\  sResult.newPrivateState \in PrivateStateType
              /\  sResult.output \in OutputType
              /\  newPrivateStateEnc \in PrivateStateEncType
              /\  currentHistorySummary \in HistorySummaryType
              /\  currentHistorySummary.anchor \in HashType
              /\  currentHistorySummary.extension \in HashType
              /\  newHistorySummary \in HistorySummaryType
              /\  newHistorySummary.anchor \in HashType
              /\  newHistorySummary.extension \in HashType
              /\  newHistorySummaryHash \in HashType
              /\  newStateHash \in HashType
              /\  newHistoryStateBinding \in HashType
              /\  newAuthenticator \in MACType
          <5>1. input \in LL2AvailableInputs
            BY <4>1
          <5>2. LL2TypeInvariant
            BY <2>1
          <5>3. QED
            BY <5>1, <5>2, LL2PerformOperationDefsTypeSafeLemma
        <4> HIDE DEF historySummaryHash, stateHash, historyStateBinding, privateState,
                sResult, newPrivateStateEnc, currentHistorySummary, newHistorySummary,
                newHistorySummaryHash, newStateHash, newHistoryStateBinding, newAuthenticator
        <4>3. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2ObservedOutputs' = LL2ObservedOutputs \union {sResult.output}
            BY <4>1 DEF sResult, privateState
          <5>3. sResult.output \in OutputType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>5. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newHistorySummaryHash,
                        newStateHash, newHistorySummary, currentHistorySummary,
                        newPrivateStateEnc, sResult, privateState
          <5>3. newAuthenticator \in MACType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>6. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM' = [ publicState |-> sResult.newPublicState,
                         privateStateEnc |-> newPrivateStateEnc,
                         historySummary |-> newHistorySummary,
                         authenticator |-> newAuthenticator ]
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, newHistorySummaryHash,
                        newStateHash, newHistorySummary, currentHistorySummary,
                        newPrivateStateEnc, sResult, privateState
          <5>2. sResult.newPublicState \in PublicStateType
            BY <4>2
          <5>3. newPrivateStateEnc \in PrivateStateEncType
            BY <4>2
          <5>4. newHistorySummary \in HistorySummaryType
            BY <4>2 DEF newHistorySummary, currentHistorySummary
          <5>5. newAuthenticator \in MACType
            BY <4>2
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5 DEF LL2UntrustedStorageType
        <4>8. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM' = [
                    historySummaryAnchor |-> LL2NVRAM.historySummaryAnchor,
                    symmetricKey |-> LL2NVRAM.symmetricKey,
                    hashBarrier |-> LL2NVRAM.hashBarrier,
                    extensionInProgress |-> TRUE ]
            BY <4>1 DEF LL2TypeInvariant, newHistorySummary
          <5>2. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>3. TRUE \in BOOLEAN
            OBVIOUS
          <5>4. QED
            BY <5>1, <5>2, <5>3 DEF LL2TrustedStorageType
        <4>9. LL2SPCR' \in HashType
          <5>1. LL2SPCR' = newHistorySummary.extension
            BY <4>1 DEF newHistorySummary, currentHistorySummary
          <5>2. newHistorySummary.extension \in HashType
            BY <4>2
          <5>3. QED
            BY <5>1, <5>2
        <4>10. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>9 DEF LL2TypeInvariant
      <3>3. CASE LL2RepeatOperation
        (*******************************************************************)
        (* For a LL2RepeatOperation action, we just walk through the       *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4>1. PICK input \in LL2AvailableInputs : LL2RepeatOperation!(input)!1
          BY <3>3 DEF LL2RepeatOperation
        <4> historySummaryHash ==
                Hash(LL2RAM.historySummary.anchor, LL2RAM.historySummary.extension)
        <4> stateHash == Hash(LL2RAM.publicState, LL2RAM.privateStateEnc)
        <4> historyStateBinding == Hash(historySummaryHash, stateHash)
        <4> newHistorySummary == Successor(LL2RAM.historySummary, input, LL2NVRAM.hashBarrier)
        <4> checkpointedHistorySummary == Checkpoint(LL2RAM.historySummary)
        <4> newCheckpointedHistorySummary ==
                Successor(checkpointedHistorySummary, input, LL2NVRAM.hashBarrier)
        <4> checkpointedNewHistorySummary == Checkpoint(newHistorySummary)
        <4> checkpointedNewCheckpointedHistorySummary ==
                Checkpoint(newCheckpointedHistorySummary)
        <4> privateState == SymmetricDecrypt(LL2NVRAM.symmetricKey, LL2RAM.privateStateEnc)
        <4> sResult == Service(LL2RAM.publicState, privateState, input)
        <4> newPrivateStateEnc ==
                SymmetricEncrypt(LL2NVRAM.symmetricKey, sResult.newPrivateState)
        <4> currentHistorySummary == [
                anchor |-> LL2NVRAM.historySummaryAnchor,
                extension |-> LL2SPCR ]
        <4> currentHistorySummaryHash == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
        <4> newStateHash == Hash(sResult.newPublicState, newPrivateStateEnc)
        <4> newHistoryStateBinding == Hash(currentHistorySummaryHash, newStateHash)
        <4> newAuthenticator == GenerateMAC(LL2NVRAM.symmetricKey, newHistoryStateBinding)
        <4>2. /\  historySummaryHash \in HashType
              /\  stateHash \in HashType
              /\  historyStateBinding \in HashType
              /\  newHistorySummary \in HistorySummaryType
              /\  newHistorySummary.anchor \in HashType
              /\  newHistorySummary.extension \in HashType
              /\  checkpointedHistorySummary \in HistorySummaryType
              /\  checkpointedHistorySummary.anchor \in HashType
              /\  checkpointedHistorySummary.extension \in HashType
              /\  newCheckpointedHistorySummary \in HistorySummaryType
              /\  newCheckpointedHistorySummary.anchor \in HashType
              /\  newCheckpointedHistorySummary.extension \in HashType
              /\  checkpointedNewHistorySummary \in HistorySummaryType
              /\  checkpointedNewHistorySummary.anchor \in HashType
              /\  checkpointedNewHistorySummary.extension \in HashType
              /\  checkpointedNewCheckpointedHistorySummary \in HistorySummaryType
              /\  checkpointedNewCheckpointedHistorySummary.anchor \in HashType
              /\  checkpointedNewCheckpointedHistorySummary.extension \in HashType
              /\  privateState \in PrivateStateType
              /\  sResult \in ServiceResultType
              /\  sResult.newPublicState \in PublicStateType
              /\  sResult.newPrivateState \in PrivateStateType
              /\  sResult.output \in OutputType
              /\  newPrivateStateEnc \in PrivateStateEncType
              /\  currentHistorySummary \in HistorySummaryType
              /\  currentHistorySummary.anchor \in HashType
              /\  currentHistorySummary.extension \in HashType
              /\  currentHistorySummaryHash \in HashType
              /\  newStateHash \in HashType
              /\  newHistoryStateBinding \in HashType
              /\  newAuthenticator \in MACType
          <5>1. input \in LL2AvailableInputs
            BY <4>1
          <5>2. LL2TypeInvariant
            BY <2>1
          <5>3. QED
            BY <5>1, <5>2, LL2RepeatOperationDefsTypeSafeLemma
        <4> HIDE DEF historySummaryHash, stateHash, historyStateBinding, newHistorySummary,
                checkpointedHistorySummary, newCheckpointedHistorySummary,
                checkpointedNewHistorySummary, checkpointedNewCheckpointedHistorySummary,
                privateState, sResult, newPrivateStateEnc, currentHistorySummary,
                currentHistorySummaryHash, newStateHash, newHistoryStateBinding,
                newAuthenticator
        <4>3. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2ObservedOutputs' = LL2ObservedOutputs \union {sResult.output}
            BY <4>1 DEF sResult, privateState
          <5>3. sResult.output \in OutputType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>5. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2ObservedAuthenticators' =
                    LL2ObservedAuthenticators \union {newAuthenticator}
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, currentHistorySummaryHash,
                        newStateHash, newPrivateStateEnc, sResult, privateState
          <5>3. newAuthenticator \in MACType
            BY <4>2
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>6. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM' = [ publicState |-> sResult.newPublicState,
                         privateStateEnc |-> newPrivateStateEnc,
                         historySummary |-> currentHistorySummary,
                         authenticator |-> newAuthenticator ]
            BY <4>1 DEF newAuthenticator, newHistoryStateBinding, currentHistorySummaryHash,
                        newStateHash, currentHistorySummary, newPrivateStateEnc, sResult,
                        privateState
          <5>2. sResult.newPublicState \in PublicStateType
            BY <4>2
          <5>3. newPrivateStateEnc \in PrivateStateEncType
            BY <4>2
          <5>4. currentHistorySummary \in HistorySummaryType
            BY <4>2
          <5>5. newAuthenticator \in MACType
            BY <4>2
          <5>6. QED
            BY <5>1, <5>2, <5>3, <5>4, <5>5 DEF LL2UntrustedStorageType
        <4>8. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>9. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>10. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>9 DEF LL2TypeInvariant
      <3>4. CASE LL2TakeCheckpoint
        (*******************************************************************)
        (* For a LL2TakeCheckpoint action, we just walk through the        *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4> newHistorySummaryAnchor == Hash(LL2NVRAM.historySummaryAnchor, LL2SPCR)
        <4>1. newHistorySummaryAnchor \in HashType
          <5>1. LL2TypeInvariant
            BY <2>1
          <5>2. QED
            BY <5>1, LL2TakeCheckpointDefsTypeSafeLemma
        <4> HIDE DEF newHistorySummaryAnchor
        <4>2. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2RAM
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2NVRAM' = [
                    historySummaryAnchor |-> newHistorySummaryAnchor,
                    symmetricKey |-> LL2NVRAM.symmetricKey,
                    hashBarrier |-> LL2NVRAM.hashBarrier,
                    extensionInProgress |-> FALSE ]
            BY <3>4 DEF LL2TakeCheckpoint, newHistorySummaryAnchor
          <5>3. newHistorySummaryAnchor \in HashType
            BY <4>1
          <5>4. FALSE \in BOOLEAN
            OBVIOUS
          <5>5. QED
            BY <5>1, <5>2, <5>3, <5>4 DEF LL2TrustedStorageType
        <4>8. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <3>4 DEF LL2TakeCheckpoint
          <5>3. QED
            BY <5>1, <5>2
        <4>9. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL2TypeInvariant
      <3>5. CASE LL2Restart
        (*******************************************************************)
        (* For a LL2Restart action, we just walk through the definitions.  *)
        (* Type safety follows directly.                                   *)
        (*******************************************************************)
        <4>1. PICK untrustedStorage \in LL2UntrustedStorageType,
                   randomSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
                   hash \in HashType :
                   LL2Restart!(untrustedStorage, randomSymmetricKey, hash)
          BY <3>5 DEF LL2Restart
        <4>2. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2RAM' = untrustedStorage
            BY <4>1
          <5>3. untrustedStorage \in LL2UntrustedStorageType
            BY <4>1
          <5>4. QED
            BY <5>1, <5>2, <5>3
        <4>7. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. LL2SPCR' \in HashType
          <5>1. LL2SPCR' = BaseHashValue
            BY <4>1
          <5>2. BaseHashValue \in HashType
            BY BaseHashValueTypeSafe
          <5>3. QED
            BY <5>1, <5>2
        <4>9. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL2TypeInvariant
      <3>6. CASE LL2ReadDisk
        (*******************************************************************)
        (* Type safety is straightforward for a LL2ReadDisk action.        *)
        (*******************************************************************)
        <4>1. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>2. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2RAM' = LL2Disk
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <3>6 DEF LL2ReadDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>8. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL2TypeInvariant
      <3>7. CASE LL2WriteDisk
        (*******************************************************************)
        (* Type safety is straightforward for a LL2WriteDisk action.       *)
        (*******************************************************************)
        <4>1. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>2. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2RAM \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. LL2Disk' = LL2RAM
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2RAM
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <3>7 DEF LL2WriteDisk
          <5>3. QED
            BY <5>1, <5>2
        <4>8. QED
          BY <4>1, <4>2, <4>3, <4>4, <4>5, <4>6, <4>7 DEF LL2TypeInvariant
      <3>8. CASE LL2CorruptRAM
        (*******************************************************************)
        (* Type safety is straightforward for a LL2CorruptRAM action.      *)
        (*******************************************************************)
        <4>1. PICK  untrustedStorage \in LL2UntrustedStorageType,
                    fakeSymmetricKey \in SymmetricKeyType \ {LL2NVRAM.symmetricKey},
                    hash \in HashType :
                    LL2CorruptRAM!(untrustedStorage, fakeSymmetricKey, hash)
          BY <3>8 DEF LL2CorruptRAM
        <4>2. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>3. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2RAM' \in LL2UntrustedStorageType
          <5>1. untrustedStorage \in LL2UntrustedStorageType
            BY <4>1
          <5>2. LL2RAM' = untrustedStorage
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. LL2SPCR' \in HashType
          <5>1. LL2SPCR \in HashType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2SPCR
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>9. QED
          BY <4>2, <4>3, <4>4, <4>5, <4>6, <4>7, <4>8 DEF LL2TypeInvariant
      <3>9. CASE LL2CorruptSPCR
        (*******************************************************************)
        (* For a LL2CorruptSPCR action, we just walk through the           *)
        (* definitions.  Type safety follows directly.                     *)
        (*******************************************************************)
        <4>1. PICK  fakeHash \in HashDomain : LL2CorruptSPCR!(fakeHash)!1
          BY <3>9 DEF LL2CorruptSPCR
        <4> newHistorySummaryExtension == Hash(LL2SPCR, fakeHash)
        <4>2. newHistorySummaryExtension \in HashType
          <5>1. LL2TypeInvariant
            BY <2>1
          <5>2. QED
            BY <5>1, LL2CorruptSPCRDefsTypeSafeLemma
        <4> HIDE DEF newHistorySummaryExtension
        <4>3. LL2AvailableInputs' \subseteq InputType
          <5>1. LL2AvailableInputs \subseteq InputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2AvailableInputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>4. LL2ObservedOutputs' \subseteq OutputType
          <5>1. LL2ObservedOutputs \subseteq OutputType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedOutputs
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>5. LL2ObservedAuthenticators' \subseteq MACType
          <5>1. LL2ObservedAuthenticators \subseteq MACType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2ObservedAuthenticators
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>6. LL2Disk' \in LL2UntrustedStorageType
          <5>1. LL2Disk \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2Disk
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>7. LL2RAM' \in LL2UntrustedStorageType
          <5>1. LL2RAM \in LL2UntrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2RAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>8. LL2NVRAM' \in LL2TrustedStorageType
          <5>1. LL2NVRAM \in LL2TrustedStorageType
            BY <2>1 DEF LL2TypeInvariant
          <5>2. UNCHANGED LL2NVRAM
            BY <4>1
          <5>3. QED
            BY <5>1, <5>2
        <4>9. LL2SPCR' \in HashType
          <5>1. newHistorySummaryExtension \in HashType
            BY <4>2
          <5>2. LL2SPCR' = newHistorySummaryExtension
            BY <4>1 DEF newHistorySummaryExtension
          <5>3. QED
            BY <5>1, <5>2
        <4>10. QED
          BY <4>3, <4>4, <4>5, <4>6, <4>7, <4>8, <4>9 DEF LL2TypeInvariant
      <3>10. QED
        BY <2>3, <3>1, <3>2, <3>3, <3>4, <3>5, <3>6, <3>7, <3>8, <3>9 DEF LL2Next
    <2>4. QED
      BY <2>1, <2>2, <2>3
  <1>3. QED
    (***********************************************************************)
    (* Using the Inv1 proof rule, the base case and the induction step     *)
    (* together imply that the invariant always holds.                     *)
    (***********************************************************************)
    <2>1. LL2TypeInvariant /\ [][LL2Next]_LL2Vars => []LL2TypeInvariant
      BY <1>2, Inv1
    <2>2. QED
      BY <2>1, <1>1 DEF LL2Spec

====
